package com.chelvc.framework.feign.interceptor;

import java.util.Collections;
import java.util.Map;

import com.chelvc.framework.base.context.ApplicationContextHolder;
import com.chelvc.framework.base.context.Session;
import com.chelvc.framework.base.context.SessionContextHolder;
import com.chelvc.framework.base.context.Using;
import com.chelvc.framework.base.util.HttpUtils;
import com.chelvc.framework.base.util.SpringUtils;
import com.chelvc.framework.common.model.Platform;
import com.chelvc.framework.common.model.Terminal;
import com.chelvc.framework.common.util.ObjectUtils;
import feign.RequestInterceptor;
import feign.RequestTemplate;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.springframework.context.ApplicationContext;
import org.springframework.web.servlet.mvc.method.annotation.RequestMappingHandlerMapping;

/**
 * Feign调用处理拦截器
 *
 * @author Woody
 * @date 2024/1/30
 */
@Slf4j
public class FeignInvokeInterceptor implements RequestInterceptor {
    private final String location;
    private final Map<String, ?> prefixes;
    private final Map<String, ?> mappings;

    @SuppressWarnings("unchecked")
    public FeignInvokeInterceptor(@NonNull ApplicationContext applicationContext) {
        this.location = "http://" + ApplicationContextHolder.getApplicationName(applicationContext.getEnvironment());
        RequestMappingHandlerMapping handlerMapping = applicationContext.getBean(RequestMappingHandlerMapping.class);
        this.prefixes = ObjectUtils.ifEmpty(handlerMapping.getPathPrefixes(), Collections::emptyMap);
        Object registry = ObjectUtils.getObjectValue(handlerMapping, "mappingRegistry");
        Map<String, ?> mappings = (Map<String, ?>) ObjectUtils.getObjectValue(registry, "urlLookup");
        this.mappings = ObjectUtils.ifEmpty(mappings, Collections::emptyMap);
    }

    /**
     * 初始化请求头
     *
     * @param template 请求模版对象实例
     */
    protected void initializeRequestHeader(@NonNull RequestTemplate template) {
        Session session = SessionContextHolder.getSession(false);
        if (session == null) {
            return;
        }
        template.header("x-real-ip", session.getHost());
        template.header("x-request-id", session.getRid());
        template.header(SessionContextHolder.HEADER_ID, (String) ObjectUtils.ifNull(session.getId(), String::valueOf));
        template.header(SessionContextHolder.HEADER_USING, ObjectUtils.ifNull(session.getUsing(), Using::name));
        template.header(SessionContextHolder.HEADER_SCOPE, session.getScope());
        template.header(SessionContextHolder.HEADER_MOBILE, session.getMobile());
        template.header(SessionContextHolder.HEADER_DEVICE, session.getDevice());
        template.header(SessionContextHolder.HEADER_CHANNEL, session.getChannel());
        template.header(SessionContextHolder.HEADER_PLATFORM,
                ObjectUtils.ifNull(session.getPlatform(), Platform::name));
        template.header(SessionContextHolder.HEADER_TERMINAL,
                ObjectUtils.ifNull(session.getTerminal(), Terminal::name));
        template.header(SessionContextHolder.HEADER_VERSION, session.getVersion());
        template.header(SessionContextHolder.HEADER_TIMESTAMP,
                (String) ObjectUtils.ifNull(session.getTimestamp(), String::valueOf));
        template.header(SessionContextHolder.HEADER_REGISTERING,
                (String) ObjectUtils.ifNull(session.getRegistering(), String::valueOf));
        template.header(SessionContextHolder.HEADER_AUTHORITIES, session.getAuthorities());
    }

    /**
     * 初始化多服务部署接口调用目标
     *
     * @param template 请求模版对象实例
     */
    protected void initializeMultipleServerTarget(@NonNull RequestTemplate template) {
        if (!SpringUtils.isMultipleServer()) {
            return;
        }

        String server = template.feignTarget().name();
        if (this.prefixes.containsKey(server)) {
            String uri = HttpUtils.uri(server, template.path());
            if (this.mappings.containsKey(uri)) {
                // 更新接口资源地址
                template.uri(uri);
            }

            // 设置本地调用地址
            template.target(this.location);
        }
    }

    @Override
    public void apply(RequestTemplate template) {
        this.initializeRequestHeader(template);
        this.initializeMultipleServerTarget(template);

        if (log.isDebugEnabled()) {
            log.debug("Feign request: {}", template);
        }
    }
}
